/***********************************************************************/
/* Reading a 123 worksheet into a C program                            */
/* ----------------------------------------                            */
/* Garry J. Vass  [72307,3311]                                         */
/***********************************************************************/
/*  This program will compile under Turbo C 1.5 and Microsoft C 5.1    */
/*  (large model) without errors.  Assumptions regarding byte ordering */
/*  obviate portability beyond the PC environment.                     */
/*                                                                     */
/***********************************************************************/
#define WK_NAME  "ERASE_ME.WK1"
/***********************************************************************/
/* Prototypes                                                          */
/*                                                                     */
/*                                                                     */
/*                                                                     */
/***********************************************************************/
#include <fcntl.h>
#include <io.h>
#include <stat.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
extern  void extract_col_row(void);
extern  void main(void);
extern  void process_123_record(void);
extern  void read_123_record(void);
extern  void test_ioresult(void);
/***********************************************************************/
/* Global variables.                                                   */
/*                                                                     */
/*                                                                     */
/*                                                                     */
/***********************************************************************/
double        dvalue;
signed   int  ioresult;
signed   int  ivalue;
signed   int  wkhandle;
unsigned char buffer[2048];
unsigned char format;
unsigned int  column;
unsigned int  length;
unsigned int  opcode;
unsigned int  row;
/***********************************************************************/
/*  This function extracts the format byte, the worksheet row, and     */
/*  the worksheet column.  The format byte is then decoded into a      */
/*  bit string.                                                        */
/*                                                                     */
/***********************************************************************/
void extract_col_row()
{
unsigned char bitstring[255];
format = buffer[0];
memmove(&column, &buffer[1], 2);
memmove(&row,    &buffer[3], 2);
(void) strcpy(bitstring, "");
/* protection bit */
if (format & 128) (void) strcat(bitstring, "1"); else (void) strcat(bitstring,"0");
(void) strcat(bitstring, "-");
/* format type bits  */
if (format &  64) (void) strcat(bitstring, "1"); else (void) strcat(bitstring,"0");
if (format &  32) (void) strcat(bitstring, "1"); else (void) strcat(bitstring,"0");
if (format &  16) (void) strcat(bitstring, "1"); else (void) strcat(bitstring,"0");
(void) strcat(bitstring, "-");
/* decimal places */
if (format &   8) (void) strcat(bitstring, "1"); else (void) strcat(bitstring,"0");
if (format &   4) (void) strcat(bitstring, "1"); else (void) strcat(bitstring,"0");
if (format &   2) (void) strcat(bitstring, "1"); else (void) strcat(bitstring,"0");
if (format &   1) (void) strcat(bitstring, "1"); else (void) strcat(bitstring,"0");
printf("\tformat:    %02X or %s\n", format, bitstring);
printf("\tlocation:    column %d, row %d\n", column, row);
}
/***********************************************************************/
/*  This function simply checks the return value from the read function*/
/*                                                                     */
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void test_ioresult()
{
if (ioresult <= 0)
	{
	printf("Read error\n");
	(void) close (wkhandle);
	exit(0);
	}
}
/***********************************************************************/
/* This function reads a single 123 record of the form <opcode>        */
/* <length> <data>.                                                    */
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void read_123_record()
{
ioresult = read(wkhandle, (unsigned char *) &opcode, 2);
test_ioresult();
ioresult = read(wkhandle, (unsigned char *) &length, 2);
test_ioresult();
/*
printf("Opcode = %04X\n", opcode);
printf("Length = %d %04X\n", length, length);
*/
if (length)
	{
	ioresult = read(wkhandle, buffer, length);
	if (ioresult != (signed int) length)
		{
		printf("Incomplete worksheet %d  %d\n", ioresult, length);
		(void) close (wkhandle);
		exit(0);
		}
	}
}
/***********************************************************************/
/*  This function provides information about various opcodes.          */
/*                                                                     */
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void process_123_record()
{
switch(opcode)
	{
	case 0x0000:
		{
		printf("Beginning of file\n");
		break;
		}
	case 0x0001:
		{
		printf("End of file\n");
		break;
		}
	case 0x0002:
		{
		printf("Calculation mode is ");
		switch (buffer[0])
			{
			case 0:
				{
				printf("manual\n");
				break;
				}
			case 0xff:
				{
				printf("automatic\n");
				break;
				}
			default:
				{
				printf("unknown\n");
				break;
				}
			}
		break;
		}
	case 0x0003:
		{
		printf("Calculation order is ");
		switch(buffer[0])
			{
            case 0:
				{
				printf("natural\n");
				break;
				}
			case 0x01:
				{
				printf("column\n");
				break;
				}
			case 0xff:
				{
				printf("row\n");
				break;
				}
			}
		break;
		}
	case 0x0006:
		{
		printf("Saved Range\n");
		memmove(&column, &buffer[0], 2);
		memmove(&row,    &buffer[2], 2);
		printf("\tFrom column %d, row %d\n", column, row);
		memmove(&column, &buffer[4], 2);
		memmove(&row,    &buffer[6], 2);
		printf("\tTo column %d, row %d\n", column, row);
		break;
		}
	case 0x0010:
		{
		printf("Formula\n");
		extract_col_row();
		memmove(&dvalue, &buffer[5], 8);
		printf("\tcurrent value:    %16.5f\n", dvalue);
		memmove(&ivalue, &buffer[13], 2);
		printf("\tformula size:    %d\n", ivalue);
		break;
		}
	case 0x000d:
		{
		printf("Integer\n");
		extract_col_row();
		memmove(&ivalue, &buffer[5], 2);
		printf("\tvalue:    %d\n", ivalue);
		break;
		}
	case 0x000e:
		{
		printf("Double\n");
		extract_col_row();
		memmove(&dvalue, &buffer[5], 8);
		printf("\tvalue:    %16.5f\n", dvalue);
		break;
		}
	case 0x000f:
		{
		printf("Label\n");
		extract_col_row();
		printf("\talignment:    %c\n", buffer[5]);
		printf("\tvalue:    %s\n", &buffer[6]);
		break;
		}
	case 0x002f:
		{
		printf("Iteration count of %d\n", buffer[0]);
		break;
		}
	case 0x0096:
		{
		printf("Cell pointer index\n");
		printf("\t%d active columns\n", (int) (length / 6));
		break;
		}
	}
}
/***********************************************************************/
/* Beginning of main.                                                  */
/*                                                                     */
/*                                                                     */
/*                                                                     */
/***********************************************************************/
void main()
{
wkhandle = open(WK_NAME, O_RDONLY | O_BINARY);
if (wkhandle <= 0)
	{
	printf("Cannot open %s\n", WK_NAME);
	exit(0);
	}
length = 1;
while (length)
	{
	read_123_record();
	process_123_record();
	}
(void) close(wkhandle);
}
